/*
 * Multicast Testing code
 *
 * (C) 2009 Linux Foundation
 *		Christoph Lameter <cl@linux-foundation.org>
 */
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include <pthread.h>
#include <linux/types.h>
#include <math.h>
#include <errno.h>
#include <ifaddrs.h>
#include <signal.h>
#include <ctype.h>

#ifndef offsetof
#define offsetof(a,b) __builtin_offsetof(a,b)
#endif

#define barrier() __asm__ __volatile__("": : :"memory")

int verbose = 0;
int multicast_loop = 0;
int port = 9002;	/* Port to listen to or send to */
unsigned localaddr;	/* Address of the NIC that we are using */

/*
 * Generic funcxtions
 */
void fatal(const char *x, ...)
{
	va_list ap;

	va_start(ap, x);
	vfprintf(stderr, x, ap);
	va_end(ap);
	exit(EXIT_FAILURE);
}

static inline long minimum(int a, int b)
{
	if (a < b)
		return a;
	return b;
}

// #define USE_TSC

#ifdef USE_TSC
#include "ll.h"
#define get_timestamp timestamp_nanoseconds

#else
static inline __u64 get_timestamp(void)
{
	struct timespec spec;

	clock_gettime(CLOCK_REALTIME, &spec);
	return spec.tv_nsec + spec.tv_sec * 1000000000LL;
}
#endif

struct in_addr toaddr(unsigned x)
{
	struct in_addr y = { x };

	return y;
}

char *ntoa(unsigned x)
{
	return inet_ntoa(toaddr(x));
}

unsigned aton(char *s)
{
	unsigned host;
	if (!inet_aton(s, (struct in_addr *)&host))
		fatal("Invalid address %s\n", s);
	return host;
}

void set_ip(struct sockaddr_in *s, unsigned host)
{
	s->sin_addr.s_addr = host;
	s->sin_family = AF_INET;
	s->sin_port = htons(port);
}

/*
 * Multicast pattern functionality
 */
struct highlow {
	unsigned char low;
	unsigned char high;
};

union pattern {
	struct highlow nibble[4];
	unsigned long long value;
};

union pattern mc_pattern = { .nibble = { {239,239}, {0,0}, {192,254}, {2,254} } };

unsigned max_groups =  15939;

int offset_into_pattern = 0;
unsigned *pattern_ips;

unsigned long calculate_ips(union pattern p)
{
	unsigned long n = 1;
	int i;

	for (i = 0; i < 4; i++)
		n *= p.nibble[i].high - p.nibble[i].low + 1;
	return n;
}

/* Create the list of valid MC addresses for patterns */
unsigned *build_pattern_array(union pattern p, unsigned n)
{
	unsigned nibble[4];
	int i;
	unsigned *pa;

	pa = malloc(sizeof(unsigned) * n);

	for (i = 0; i < 4; i++)
		nibble[i] = p.nibble[i].low;

	n = 0;
	do {
		pa[n++] = htonl((nibble[0] << 24) + (nibble[1] << 16) + (nibble[2] << 8) + nibble[3]);

		i = 3;
		while (i >= 0) {
			nibble[i]++;
			if (nibble[i] <= p.nibble[i].high)
				break;
			nibble[i] = p.nibble[i].low;
			i--;
		}
	} while (i >= 0);

	return pa;
}

int parse_pattern(char *s)
{
	int i;

	if (isdigit(*s)) {
		for (i = 0; i < 4; i++) {
			if (i && *s++ != '.') {
				printf(". expected in pattern\n");
				return 1;
			}

			mc_pattern.nibble[i].low = mc_pattern.nibble[i].high = strtoul(s, &s, 10);
			if (*s == '-') {
				s++;
				mc_pattern.nibble[i].high = strtoul(s, &s, 10);
			}
		}

		max_groups = calculate_ips(mc_pattern);
		if (max_groups < 1 || max_groups > 32000) {
			printf("Invalid pattern\n");
			return 1;
		}
		return 0;
	}

	printf("Unknown pattern '%s'\n",s);
	return 1;
}

char *print_pattern(union pattern p)
{
	static char buffer[20];
	char *s = buffer;
	int i;

	for (i = 0; i < 4; i++) {
		if (p.nibble[i].low == p.nibble[i].high)
			s += sprintf(s, "%d", p.nibble[i].low);
		else
			s += sprintf(s, "%d-%d", p.nibble[i].low, p.nibble[i].high);
		if (i < 3)
			*s++ = '.';
	}
	*s = 0;
	return buffer;
}

/*
 * Calculate the Nth IP address starting at a base ip address.
 */
unsigned inline multicast_group(int n)
{
	return pattern_ips[n + offset_into_pattern];
}

/*
 * Open socket
 */
int open_sock(int bind_socket, char *interface, int ttl)
{
	int sock;
	int err;
	int yes = 1;
	int no = 0;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock < 0) {
		perror("socket");
		fatal("socket error\n");
	}

	if (bind_socket) {
		struct sockaddr_in inad;

		err = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
		if (err) {
			perror("reuseaddr");
			fatal("reuseaddr error\n");
		}

		/* Bind to the port .... */
		set_ip(&inad, INADDR_ANY);
		if (bind(sock, &inad, sizeof(inad))) {
			perror("bind");
			fatal("bind erro\nr");
		}
	}

	if (interface) {
		err = setsockopt(sock, SOL_SOCKET, SO_BINDTODEVICE, interface, strlen(interface) + 1);

		if (err) {
			perror("bindtodevice");
			fatal("bindtodevice error\n");
		}
	}

	if (ttl >= 0) {
		err = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl));
		if (err) {
			perror("setsockopt IP_MULTICAST_TTL\n");
			fatal("setsockopt error\n");
		}
	}

#ifdef IP_MULTICAST_ALL
	setsockopt(sock, IPPROTO_IP, IP_MULTICAST_ALL, &no, sizeof(no));
#endif

	/*
	 * We are using two sockets. If we are not switching off loopback then everything
	 * we send will come back to us.
	 */
	if (!multicast_loop) {
		err = setsockopt(sock, IPPROTO_IP, IP_MULTICAST_LOOP, &no, sizeof(no));
		if (err) {
			perror("setsockopt IP_MULTICAST_LOOP\n");
			fatal("setsockopt error\n");
		}
	}

	return sock;
}

/*
 * Join a multicast group
 */
int mcast_join(int sock, unsigned group)
{

	struct ip_mreq mreq;
	struct sockaddr_in addr;

	set_ip(&addr, group);

	mreq.imr_multiaddr.s_addr = group;
	mreq.imr_interface.s_addr = localaddr;
	if (verbose > 1)
		printf("Joining mcast group %s\n", ntoa(group));

	if (setsockopt(sock, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
		perror("setsockopt(mcast join)");
		fatal("Join error\n");
	}

	return 0;
}

int mcast_leave(int sock, unsigned group)
{
	struct ip_mreq mreq;
	struct sockaddr_in addr;

	set_ip(&addr, group);

	mreq.imr_multiaddr.s_addr = group;
	mreq.imr_interface.s_addr = localaddr;
	if (verbose > 1)
		printf("Leaving mcast group %s\n", ntoa(group));

	if (setsockopt(sock, IPPROTO_IP, IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq)) < 0) {
		perror("setsockopt(mcast_leave)");
		fatal("MCast Leave error\n");
	}

	return 0;
}

enum control { C_NOOP, C_DATA, C_TESTEND, C_NEW_MC_GROUP, C_RM_MC_GROUP,
	C_DUMP_STATISTICS, C_LAT_REQUEST, C_LAT_RESPONSE, C_NACK_MC, C_ZAP };

char *control_text[] = {
	"No Operation",
	"Data",
	"End of test",
	"New Multicast Group",
	"Remove Multicast Group",
	"Dump statistics",
	"Startup",
	"Latency Data request",
	"Latency Data response",
	"Multicast Nack",
	"Counter Zap"
};

#define MAGIC 0xFEEDABCD

struct message {
	unsigned magic;
	unsigned to;		/* Sender */
	unsigned from;		/* MC or Unicast target */
	unsigned a;		/* Multicast group this is regarding or seqid */
	__u64 b;	/* Generic parameter. Usually timestamp */
	enum control control;	/* Control token */
	char payload[];
};

int ttl = 20;			/* Set ttl */
unsigned long samples = 0;	/* How many samples to send on each MC group */
int synchronize_systems = 0;	/* Number of systems to syncrhonize */
int randomize = 0;		/* Randomize send */
int burst = 1;			/* Burst size */
enum { STATS_CONTINUOUS, STATS_SINGLE_SHOT} stats_mode = STATS_CONTINUOUS;
int aggregate = 0;		/* Use OS to coalesce packets */
int length = 300;
int histogram = 0;
int mc_group_display = 0;
int errors = 0;
int terminate_remote = 0;	/* Terminate remote listener after test */
__u64 lat_interval = 1000000000LL; /* latency interval 1s */
__u64 min_lat_interval = 1000000L; /* Min time between latency probes */
__u64 last_lat;			/* Last Latency probe */

int number_of_multicast_groups = 0;

unsigned bindaddr;		/* IP address to bind to */
char *interface;		/* Interface to use */
unsigned control_channel;	/* Multicast control channel. Defaults to 239.0.192.1 */

int rate = 10;					/* Limit rate */
struct timespec wait_time = { 0, 0 };		/* Pause between packets */
__u64 receiver_start;
__u64 sender_start;
__u64 sample_start;
__u64 stat_interval = 10000000000LL;	/* 10 seconds */

int corrupted = 0;

/*
 * Format a message of length bytes
 */
void init_message(struct message *m, int length)
{
	int i;

	m->magic = MAGIC;
	m->from = localaddr;

	for (i = 0; i < length - offsetof(struct message, payload); i++)
		m->payload[i] = i;
}

/* Send UDP message to a target address with a control token and a sequence id */
void send_data_messages(int sock, struct message *m,
			unsigned to, unsigned seqid, int burst)
{
	struct sockaddr_in addr;
	int i;

	if (verbose > 3)
		 printf("Sending messages to %s, seqid=%d burst=%d\n",
			ntoa(to), seqid, burst);

	set_ip(&addr, to);

	m->to = to;
	m->control = C_DATA;
	for (i = 0; i < burst; i++) {
		int flags;

		m->a = seqid;
		m->b = get_timestamp();

		if (aggregate && i < burst - 1)
			flags = MSG_MORE;
		else
			flags = 0;

		if (sendto(sock, m, length, 0, &addr, sizeof(addr)) < 0) {
			perror("sendto");
			fatal("Socket Send error\n");
		}
		seqid++;
	}
}

/* Send UDP message to a target address without a payload */
void send_message(int sock, struct message *m, unsigned to, enum control c,
		unsigned a, __u64 b)
{
	struct sockaddr_in addr;

	if (verbose)
		 printf("Sending short message to %s, control=%s from=%s\n",
			strdupa(ntoa(to)), control_text[c],strdupa(ntoa(m->from)));

	set_ip(&addr, to);

	m->to = to;
	m->control = c;
	m->a = a;
	m->b = b;
	if (sendto(sock, m, offsetof(struct message, payload), 0, &addr, sizeof(addr)) < 0) {
		perror("sendto");
		fatal("Socket short Send error\n");
	}

}

/*
 * Send a control message for all multicast groups that we send on
 */
void mc_group_notify(int sock, enum control c)
{
	struct message *m = alloca(length);

	init_message(m, length);

	send_message(sock, m, control_channel, c,
		(offset_into_pattern << 16) + number_of_multicast_groups,
		mc_pattern.value);
}

/*
 * Statistics subsystem
 */
#define MAX_US 100000L

int mc_channels;
int hosts;

struct statistics {
	unsigned mc_addr;	/* if mc_addr == host then this is a host record */
	unsigned host;
	unsigned seqid;
	struct statistics *next;
	unsigned long min,max;
	unsigned long samples;
	__u64 sum;
	double sqsum;
	unsigned outliers, sequence_errors;
	unsigned long bytes;
	unsigned long messages;
	__u64 last_lat, start_time;
	unsigned long messages_last;
	__u64 timestamp_last;
	char *channel;		/* Name for pretty printing */
	unsigned histogram[MAX_US];
};

#define STAT_HASH_SIZE 1024

struct statistics *stat[STAT_HASH_SIZE];

int stathash(unsigned mc_addr)
{
	return mc_addr % STAT_HASH_SIZE;
}

struct statistics *find_data(unsigned mc_addr)
{
	unsigned hash = stathash(mc_addr);
	struct statistics *s = stat[hash];

	while (s && s->mc_addr != mc_addr) {
		s = s->next;
	}
	return s;
}

void mc_join(int sock, unsigned group, int host)
{
	if (!find_data(group)) {
		struct statistics *new = calloc(1, sizeof(struct statistics));
		int hash;

		new->mc_addr = group;
		new->host = host;
		new->seqid = 0;
		new->start_time = get_timestamp();
		new->channel = strdup(ntoa(group));
		mc_channels++;

		hash = stathash(group);
		new->next = stat[hash];
		stat[hash] = new;
		mcast_join(sock, group);
	}
}

void mc_leave(int sock, unsigned group)
{
	int hash = stathash(group);
	struct statistics *s = stat[hash];
	struct statistics *p = NULL;

	while (s && s->mc_addr != group) {
		p = s;
		s = s->next;
	}

	if (!s) {
		printf("mc_leave(): Cannot find MC addr %s\n", ntoa(group));
		return;
	}

	if (p)
		p->next = s->next;
	else
		stat[hash] = s->next;

	mcast_leave(sock, group);
	free(s->channel);
	free(s);
}

int hist[MAX_US];

void histo(int samples)
{
	int i, count, scale, max;

	printf("\nMicrosec\tPercent\tSamples\n");

	/* Conflate array to have less than 20 entries */
	count = 999999;
	scale = 1;
	max = MAX_US;

	count = 0;
	for (i = 0; i < max; i++)
		if (hist[i])
			count++;

	while (count > 20) {
		int newmax = max / 2;

		/* Conflate array */
		for (i = 0; i < newmax; i++)
			hist[i] = hist[ 2* i ]  + hist [ 2*i + 1 ];

		if (max % 2)
			hist[newmax - 1] += hist[max - 1];

		max = newmax;
		scale = scale * 2;
		/* Recount */
		count = 0;
		for (i = 0; i < max; i++)
			if (hist[i])
				count++;
	}

	for(i=0; i < max; i++)
		if (hist[i])
			printf("%4.0f-%.0f\t%4.0f%%\t%8u\n",
				(double)i * scale,
				(double)(i + 1) * scale,
				hist[i] * (double)100.0 / samples,
				hist[i]);
}


void add_host(unsigned host)
{
	if (!find_data(host)) {
		struct statistics *new = calloc(1, sizeof(struct statistics));
		int hash;

		new->mc_addr = host;
		new->host = host;
		new->seqid = 0;
		new->start_time = get_timestamp();
		new->channel = strdup(ntoa(host));
		hosts++;

		hash = stathash(host);
		new->next = stat[hash];
		stat[hash] = new;
	}
}

void remove_host(unsigned host)
{
	int hash = stathash(host);
	struct statistics *s = stat[hash];
	struct statistics *p = NULL;

	while (s && s->mc_addr != host) {
		p = s;
		s = s->next;
	}

	if (!s) {
		printf("remove_host(): Cannot find host addr %s\n", ntoa(host));
		return;
	}

	if (p)
		p->next = s->next;
	else
		stat[hash] = s->next;

	free(s->channel);
	free(s);
	hosts--;
}

struct statistics a;

void print_statistics(void)
{
	unsigned i;
	struct statistics *s;
	int number_stats = 0;
	double total_kbytes = 0;
	double sstddev = 0;
	double savg = 0;
	unsigned long tmin = 0, tmax = 9;
	unsigned long total_sequence_errors = 0, total_messages = 0, total_outliers = 0, total_samples = 0;
	int idle =  0;
	__u64 ts = get_timestamp();
	double runtime = (ts - sender_start) / 1000000000.0;
	double sampletime = (ts - sample_start) / 1000000000.0;

	if (histogram)
		memset(hist, 0, sizeof(hist));

	if (mc_group_display) {
		/* Loop over all received channel and print statistics */
		printf("\n----------------Summary---t=%.0fsec----------------------------------------------------------\n", runtime);
		printf("Messages Errors Msg/sec   Min/us  Avg/us  Max/us Stddev kbytes/sec   Group      Source    Samples\n");
		printf("-------------------------------------------------------------------------------------------------\n");
	}

	for (i = 0 ; i < STAT_HASH_SIZE; i++)
		for (s = stat[i]; s; s = s->next)
		if (s->mc_addr != s->host) {

		__u64 run_time;
		unsigned messages_in_interval;
		unsigned long kbytes;
		double run_seconds;
		double avg, stddev;

		memcpy(&a, s, sizeof(a));
		s->sequence_errors = s->samples = s->bytes = s->sum = s->min = s->max = 0;
		s->messages_last = a.messages;
		s->timestamp_last = ts;
		s->sqsum = 0;

		total_sequence_errors += a.sequence_errors;
		total_samples += a.samples;

		messages_in_interval = a.messages - a.messages_last;
		run_time = ts - a.timestamp_last;
		run_seconds = (double)run_time / (double)1000000000.0;
		if (!messages_in_interval)
			idle++;

		total_messages += messages_in_interval;
		kbytes = (a.bytes + 500) / 1000;
		total_kbytes += kbytes;

		if (a.samples) {
			avg = ((double)a.sum) / a.samples / 1000.0;
			savg += avg;
			if (a.max > tmax)
				tmax = a.max;
			if (a.min < tmin || !tmin)
				tmin = a.min;
			stddev = sqrt(a.sqsum / a.samples / 1000000.0 - avg * avg);
			sstddev += stddev;
			if (histogram) {
				int i;
				for(i = 0 ; i < MAX_US; i++)
					hist[i] += a.histogram[i];
			}
			number_stats++;
		}

		if (mc_group_display && a.messages) {
			printf("%8d %6d %7.0f ", messages_in_interval, a.sequence_errors, messages_in_interval / run_seconds);
			printf("%7.2f ", a.min / 1000.0);
			printf("%7.2f ", avg);
			printf("%7.2f ", a.max / 1000.0);
			printf("%7.2f ", stddev);
			printf("%7.1f ",  kbytes / run_seconds);
			printf("%12s %12s %6ld\n", s->channel, inet_ntoa(toaddr(a.host)), a.samples);
		}
	}

	if (total_messages) {
		if (mc_group_display)
			/* Sum it all up and print more statistics */
			printf("----Totals------------------------------Groups=%d----------------Idle------Samples----\n",
					mc_channels);
		else {
			static int first = 1;

			if (first || stats_mode == STATS_SINGLE_SHOT) {
				printf("\n   Time TotalMsg SeqErr Msg/Sec   Min/us  Avg/us  Max/us StdDev  Kbytes Idle  Samples\n");
				first = 0;
			}
			if (total_outliers || corrupted)
				printf("Outliers-%ld. Corrupted %d\n", total_outliers, corrupted);

			printf("%7.0f ", runtime);
		}
	}

	if (total_messages) {
		printf("%8ld %6ld %7.0f ", total_messages, total_sequence_errors, total_messages / sampletime);
		printf("%7.2f ", tmin / 1000.0);
		printf("%7.2f ", savg / number_stats);
		printf("%7.2f ", tmax / 1000.0);
		printf("%7.2f ", sstddev / number_stats);
		printf("%7.1f ", total_kbytes);
		printf("%4d ", idle);
		printf("%8ld\n", total_samples);
		if (histogram)
			histo(total_samples);
		if (stats_mode == STATS_SINGLE_SHOT)
			printf("\n");
	}

	corrupted = 0;
	sample_start = get_timestamp();
}

void clear_statistics(void)
{
	unsigned i;
	struct statistics *s;

	for (i = 0 ; i < STAT_HASH_SIZE; i++)
		for (s = stat[i]; s; s = s->next) {

		s->sequence_errors = s->samples = s->bytes = s->sum = s->min = s->max = 0;
		s->messages_last = 0;
		s->timestamp_last = 0;
		s->sqsum = 0;
	}

	corrupted = 0;
	sample_start = get_timestamp();
}

/*
 * Sender thread
 *
 * Thread sends messages to all MC groups in a loop and then delays to get the proper timing.
 *
 * Statistics are generated in regular intervals. MC groups are announced when statistics are
 * generated.
 */
enum { SENDER_OFF, SENDER_SYNC, SENDER_RUNNING  } sender_mode;
enum { RECEIVER_OFF, RECEIVER_ON } receiver_mode;
enum { SENDER_NEED_NO_SYNC, SENDER_NEED_SYNC, SENDER_NEED_STOP }
		sender_need_sync = SENDER_SYNC;

int send_sock;

void sigint(int x)
{
	sender_need_sync = SENDER_NEED_STOP;
	printf("\nUnsubscribing from multicast groups\n");
	mc_group_notify(send_sock, C_RM_MC_GROUP);
	exit(1);
}

void *sender_thread(void *dummy)
{
	int i;
	struct message *m = alloca(length);
	unsigned *seqids = calloc(sizeof(unsigned), number_of_multicast_groups);
	unsigned pass = 0;
	__u64 ts;
	__u64 next_stats;
	__u64 next;
	__u64 send_interval;
	struct sigaction sigint_old;
	struct sigaction sigint_new;

	send_sock = open_sock(0, interface, ttl);
	init_message(m, length);

	/*
	 * Start sending
	 */
	for (i = 0; i < number_of_multicast_groups; i++)
		seqids[i] = 1;

	sigint_new.sa_handler = sigint;
	sigemptyset(&sigint_new.sa_mask);
	sigint_new.sa_flags = SA_RESETHAND;
	sigaction(SIGINT, &sigint_new, &sigint_old);

resync:
	if (sender_need_sync == SENDER_NEED_STOP)
		goto out;

	sender_mode = SENDER_SYNC;
	sender_need_sync = SENDER_NEED_NO_SYNC;

	while (receiver_mode != RECEIVER_ON)
		barrier();

	if (max_groups < offset_into_pattern + number_of_multicast_groups)
		fatal("Number of multicast groups exhausted\n");

	do {

		if (verbose)
			printf("Sender: Announcing %d multicast groups pattern %s offset %d burst %d random %d\n",
				number_of_multicast_groups,
				print_pattern(mc_pattern),
				offset_into_pattern, burst, randomize);

		mc_group_notify(send_sock, C_NEW_MC_GROUP);

		if (synchronize_systems && !sender_need_sync && hosts + 1 < synchronize_systems) {

			if (verbose)
				printf("Waiting on %d systems\n", synchronize_systems - hosts - 1);

			sender_start = get_timestamp();
			while (hosts + 1 < synchronize_systems && get_timestamp() <
				sender_start + 100000000LL && !sender_need_sync)
				barrier();

		} else
			sleep(1);

		/*
		 * Other entities may nack our proposed MC group setup. Then the offset_into_pattern will have
		 * changed. Best to restart then.
		 */
		if (sender_need_sync)
			goto resync;

	} while (synchronize_systems && hosts + 1 < synchronize_systems);

	if (samples > 0)
		send_message(send_sock, m, control_channel, C_ZAP, STATS_SINGLE_SHOT, 0);

	if (number_of_multicast_groups)
		printf("Sender: Sending %d msgs/ch/sec on %d channels. Probe interval=%g-%g sec.\n",
			rate, number_of_multicast_groups,
			min_lat_interval / 1000000000.0,
			lat_interval / 1000000000.0);

	/* start sending */
	sender_start = get_timestamp();
	sample_start = sender_start;
	next_stats = sender_start + stat_interval;
	sender_mode = SENDER_RUNNING;

	send_interval = 1000000000LL / rate;

	for (next = get_timestamp() + send_interval; !sender_need_sync;
							next += send_interval) {
		long long wait_ns;

		ts = get_timestamp();

		wait_ns = next - ts;
		if (verbose > 4)
			printf("ts=%lld next=%lld wait_ns=%lld\n", ts, next, wait_ns);
		if (wait_ns > 0) {
			wait_time.tv_nsec = wait_ns;
			nanosleep(&wait_time, NULL);
		}

#if 0
		 else if (wait_ns < 500000000LL) {
			static int behind = 0;
			/* More than half a second behind ! */
			printf("Sender: Unable to send at requested rate. Lowering rate.\n");
			send_interval +=  1000 << behind;
			behind++;
			next = ts + send_interval;
		}
#endif

		for (i = 0; i < number_of_multicast_groups && !sender_need_sync; i++) {
			unsigned mc_group;
			unsigned msgs;

			if (randomize)
				mc_group = random() % number_of_multicast_groups;
			else
				mc_group = i;

			if (burst < 0)
				msgs = 1 + random() % (-burst -1);
			else
				msgs = burst;

			send_data_messages(send_sock, m, multicast_group(mc_group), seqids[mc_group], msgs);
			seqids[mc_group] += msgs;
		}

		if (ts > next_stats) {
			if (stats_mode == STATS_CONTINUOUS)
				print_statistics();
			next_stats = ts + stat_interval;

			/* Reannounce the multicast groups that we carry */
			mc_group_notify(send_sock, C_NEW_MC_GROUP);
		}
		pass++;
		if (pass == samples)
			break;
	}
	if (sender_need_sync)
		goto resync;

	if (samples > 0) {
		sleep(1);
		send_message(send_sock, m, control_channel, C_DUMP_STATISTICS, STATS_CONTINUOUS, 0);
	}

out:
	/*
	 * Close sender
	 */
	sigaction(SIGINT, &sigint_old, NULL);
	mc_group_notify(send_sock, C_RM_MC_GROUP);
	sender_mode = SENDER_OFF;
	free(seqids);
	if (terminate_remote)
		send_message(send_sock, m, control_channel, C_TESTEND, pass, 0);
	close(send_sock);
	printf("Done.\n");
	exit(0);
}


void start_sender(void)
{
	pthread_attr_t attr;
	pthread_t sender_thread_t;
	int rc;

	sender_mode = SENDER_OFF;

	pattern_ips = build_pattern_array(mc_pattern, max_groups);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	rc = pthread_create(&sender_thread_t, &attr, sender_thread, NULL);
        if (rc < 0)
		fatal("Cannot create sender thread\n");
	if (verbose)
		printf("Started sender\n");
}

void stop_sender(void)
{
	if (verbose)
		printf("Stopping sender\n");

	if (sender_mode == SENDER_OFF)
		return;

	while (sender_mode != SENDER_RUNNING)
		barrier();

	sender_need_sync = SENDER_NEED_STOP;

	while (sender_mode != SENDER_OFF)
		barrier();
}


/*
 * Receiver thread
 *
 * Simply waits for traffic to come in on a port and then processes the data.
 */

/* Special messaging structure for replies */
struct message *reply;

/* Receiver has requested latency information for a message and provided a timestamp */
void lat_request(int sock, struct message *m, __u64 ts)
{
	unsigned group = m->a;
	__u64 timestamp = m->b;

	if (verbose > 1)
		printf("Latency request from %s\n", ntoa(m->from));

	send_message(sock, reply, m->from, C_LAT_RESPONSE, group, (ts - timestamp) / 2);
}

void lat_response(int sock, struct message *m)
{
	unsigned group = m->a;
	unsigned long diff = m->b;
	struct statistics *s = find_data(group);

	if (verbose)
		printf("Latency response from %s diff %ld nsec\n", ntoa(m->from), diff);

	if (!s) {
		printf("Lat response for MC address %s origin %s which we do not subscribe to.\n",
			strdupa(ntoa(group)), strdupa(ntoa(m->from)));
		return;
	}


	/* Record data */
	if (diff < 0 || (histogram && diff > MAX_US * 100)) {
		if (s->outliers < 10) {
			if (errors) {
				printf("Outlier #%d: Latency=%7.2fusecs\n",
					s->outliers + 1, diff / 1000.0);
				if (s->outliers == 9)
					printf("Too many outliers.\n");
			}
		}
		s->outliers++;
	} else {
		int udiff;

		if (diff < s->min || s->min == 0)
			s->min = diff;
		if (diff > s->max)
			s->max = diff;

		udiff = diff / 1000;
		if (histogram)
			s->histogram[ udiff]++;
		s->samples++;
		s->sum += diff;
		s->sqsum += (double)diff * diff;
	}
}

void handle_data(int sock, struct message *m, __u64 ts, unsigned length)
{
	struct statistics *s = find_data(m->to);
	unsigned seqid = m->a;
	__u64 timestamp = m->b;

	if (!s) {
		printf("Data for MC address %s origin %s which we do not subscribe to.\n",
			strdupa(ntoa(m->to)), strdupa(ntoa(m->from)));
		return;
	}

	if (seqid != s->seqid) {
		if (s->seqid && seqid != 1) {
			if (errors)
				printf("*** Channel %s: Message out of sequence. Expected %d received %d length=%d\n",
					s->channel, s->seqid, seqid, length);
			s->sequence_errors++;
		}
		s->seqid = seqid;
	}

	if (verbose > 2)
		printf("Channel %s: %d Received data (from %s). length=%d\n",
			s->channel, seqid,
			strdupa(ntoa(m->from)),
			length);

	s->messages++;
	s->seqid++;
	s->bytes += length;
	if (s->last_lat + lat_interval < ts && last_lat + min_lat_interval < ts) {
		send_message(sock, reply, m->from, C_LAT_REQUEST, m->to, timestamp);
		s->last_lat = last_lat = ts;
	}
}

void new_mc_groups(int sock, struct message *m)
{
	union pattern p;;
	unsigned offset;
	unsigned *ips;
	unsigned groups;
	unsigned max_groups;
	int i,j;

	p.value = m->b;
	offset = m->a >> 16;
	groups = m->a & 65535;

	max_groups = calculate_ips(p);
	ips = build_pattern_array(p, max_groups);

	if (find_data(m->from))
		/* We already have informatin from that host */
		goto out;

	/* Check for conflicts */
	if (!multicast_loop) {
		for (i = offset; i < offset + groups; i++) {
			for (j = offset_into_pattern; j < offset_into_pattern + number_of_multicast_groups; j++)
				if (ips[i] == pattern_ips[j]) {
					if (verbose)
						printf("MC group NACK to %s due to overlap with our MC address %s.\n",
							strdupa(ntoa(m->from)), strdupa(ntoa(ips[i])));
					send_message(sock, reply, m->from, C_NACK_MC, offset + groups, p.value);
					goto out;
				}
		}
	}

	printf("Receiver: Subscribing to %d MC addresses %s offset %d origin %s\n",
		groups,
		print_pattern(p),
		offset, strdupa(ntoa(m->from)));

	sender_need_sync = SENDER_NEED_SYNC;
	for (i = 0; i < groups; i++) {
		mc_join(sock, ips [offset + i], m->from);
		if ((i % 100) == 0 && i)
			printf("%dth mc group = %s\n", i, ntoa(ips[offset + i]));
	}

	if (i > 100)
		printf("Finished subscribing to %d mc groups\n", i);

	add_host(m->from);

out:
	free(ips);
}

void remove_mc_groups(int sock, struct message *m)
{
	union pattern p;
	unsigned offset;
	unsigned groups;
	unsigned max_groups;
	unsigned *ips;
	int i;

	p.value = m->b;
	offset = m->a >> 16;
	groups = m->a & 65535;

	max_groups = calculate_ips(p);
	ips = build_pattern_array(p, max_groups);

	printf("Unsubscribing from %d multicast addresses origin %s.\n",
		groups, ntoa(m->from));

	for (i = 0; i < groups; i++) {
		mc_leave(sock, ips [offset + i]);
		if ((i % 100) == 0 && i)
			printf("%dth mc group =%s\n", i, ntoa(ips[offset + i]));
	}

	if (i > 100)
		printf("Finished unsubscribing from %d mc groups.\n", i);

	free(ips);
	remove_host(m->from);
}

void receiver_loop(void)
{
	struct message *m = alloca(length);
	int sock;
	__u64 ts;
	struct sockaddr_in inaddr;
	unsigned inaddrlen = sizeof(struct sockaddr_in);
	unsigned long total_messages = 0;

	/* initialize variables */
	corrupted = 0;
	sock = open_sock(1, interface, ttl);

	set_ip(&inaddr, bindaddr);
	/* Join control channel */
	mcast_join(sock, control_channel);

	printf("Receiver: Listening to control channel %s\n", ntoa(control_channel));
	receiver_start = get_timestamp();
	receiver_mode = RECEIVER_ON;

	do {
		int nbytes;

		nbytes = recvfrom(sock, m, length, 0, &inaddr, &inaddrlen);
		ts = get_timestamp();
		if (nbytes < 0) {
			perror("recvfrom");
			break;
		}

		if (verbose > 2)
			printf("Receiver_loop: %d bytes cmd=%s from=%s to=%s\n",
				nbytes, control_text[m->control],strdupa(ntoa(m->from)), strdupa(ntoa(m->to)));

		if (m->magic != MAGIC) {
			printf("** Corrupted packet ignored MAGIC=%x to=%s from=%s\n",
				m->magic, strdupa(ntoa(m->to)), strdupa(ntoa(m->from)));
			corrupted++;
			continue;
		}

		switch (m->control) {

		case C_NOOP:
			/* Ignore */
			break;

		case C_DATA:
			handle_data(sock, m, ts, nbytes);
			break;

		case C_LAT_REQUEST:
			lat_request(sock, m, ts);
			break;

		case C_LAT_RESPONSE:
			lat_response(sock, m);
			break;

		case C_NEW_MC_GROUP:
			new_mc_groups(sock, m);
			break;

		case C_RM_MC_GROUP:
			remove_mc_groups(sock, m);
			break;

		case C_NACK_MC:
			if (m->a > offset_into_pattern)
				offset_into_pattern = m->a;
			if (verbose)
				printf("Receiver: MC group NACK received. Offset is %d\n", offset_into_pattern);
			break;

		case C_DUMP_STATISTICS:
			print_statistics();
			stats_mode = m->a;
			sample_start = get_timestamp();
			break;

		case C_ZAP:
			clear_statistics();
			stats_mode = m->a;
			sample_start = get_timestamp();
			break;

		case C_TESTEND:
			goto out;

		}
		total_messages++;
	} while (1);
out:
	receiver_mode = RECEIVER_OFF;
	close(sock);
}


void find_local_addr(void)
{
	struct ifaddrs *ifaddr, *ifa;
	int family;
	int nr_match = 0;

	if (getifaddrs(&ifaddr) == -1 ) {
		perror("getifaddr");
		fatal("getofaddr failure\n");
	}

	for (ifa = ifaddr; ifa; ifa = ifa->ifa_next) {
		if (!ifa->ifa_addr)
			continue;

		family = ifa->ifa_addr->sa_family;

		if (family != AF_INET)
			continue;

		if (!interface) {
			if  (strcmp(ifa->ifa_name, "lo") == 0)
				continue;
		} else if (strcmp(ifa->ifa_name, interface))
			continue;

		if (!localaddr)
			localaddr = ((struct sockaddr_in *)ifa->ifa_addr)->sin_addr.s_addr;

		nr_match++;
	}
	freeifaddrs(ifaddr);

	if (nr_match > 1)
		printf("WARNING: Multiple active ethernet devices. Using local address %s\n", ntoa(localaddr));

	if (nr_match == 0)
		fatal("Interface %s not found\n", interface);

	if (verbose)
		printf("Using interface %s addr=%s\n", interface, ntoa(localaddr));
}

/*
 * Command line processing
 */
void usage(void)
{
	printf("mcast: Multicast stress test 8/10/2009\n"
		"(C) 2009 Linux Foundation Christoph Lameter <cl@linux-foundation.org>\n\n"
		"mcasttest {<options>} <number of multicast channels to create>\n\n"
		"-b<interface>|--bind <interface>  Bind sockets to a specific interface\n"
		"-c<ip>|--control <ip>             Multicast control channel [239.0.192.1]\n"
		"-d|--details                      Print histogram\n"
		"-e|--errors                       Show details for sequence errors\n"
		"-h|--help                         Print usage information\n"
		"-i<sec>|--interval <sec>          Statistics display interval[10]\n"
		"-j<a>.<b>.<c>[-<f>].<d>[-<e>]     Pattern for generating MC addresses[238.0.192-255.2-254]\n"
		"-l<length>|--length <length>      Set length of packet[300]\n"
		"-n<mcgroups>|--number <mcgroups>  The number of multicast groups to use[0]\n"
		"-m|--multicast                    Display a list of multicast group stats\n"
		"-o<offset>|--offset <offset>      Offset into pattern for the first MC group\n"
		"-p<port>|--port <port>            Port to use [9002]\n"
		"-r<packets>|--rate <packets>      Packets to send per second(per mc group)[10]\n"
		"-s<samples>|--samples <samples>   Samples to send on each MC group[0]\n"
		"-t<ttl>|--ttl <ttl>               Set time to live[20]\n"
		"-v|--verbose                      Increase detail of logging\n"
		"-A|--aggregate                    Use OS to coalesce multiple packets\n"
		"-B<number>| --burst <number>      Send bursts of messages on a single MC group\n"
		"-L|--loop                         Multicast loopback (required for localhost)\n"
		"-R|--random                       Avoid sequential send sequence\n"
		"-S<systems>|--synchronize <sys>   Synchronized start with N systems\n"
		"-X|--exit                         Terminate all mcast instances when test finishes\n"
	);
}

struct option opts[] = {
	{ "bind", 1, NULL, 'b' },
	{ "control", 1, NULL, 'c' },
	{ "details", 0, NULL, 'd' },
	{ "errors", 1, NULL, 'e' },
	{ "help", 0, NULL, 'h' },
	{ "join", 1, NULL, 'j' },
	{ "length", 1, NULL, 'l' },
	{ "number", 1, NULL, 'n' },
	{ "offset", 1, NULL, 'o' },
	{ "port", 1, NULL, 'p' },
	{ "rate", 1, NULL, 'r' },
	{ "samples", 1, NULL, 's' },
	{ "ttl", 1, NULL, 't' },
	{ "verbose", 0, NULL, 'v' },
	{ "aggregate", 0, NULL, 'a' },
	{ "burst", 1, NULL, 'B' },
	{ "loop", 0, NULL, 'L' },
	{ "random", 0, NULL, 'R' },
	{ "synchronize", 1, NULL, 'S' },
	{ "terminate", 0, NULL, 'X' },
	{ NULL, 0, NULL, 0 }
};

int main(int argc, char *argv[])
{
	int c;

	control_channel = aton("239.0.192.1");

	while ((c = getopt_long(argc, argv, "b:c:dehi:j:l:mn:o:p:r:s:t:vAB:LRS:X",
							opts, NULL)) != -1)
	switch (c) {
	case 'b' :
		interface = strdup(optarg);
		break;

	case 'c' :
		control_channel = aton(optarg);
		break;

	case 'd':
		histogram = 1;
		break;

	case 'e':
		errors = 1;
		break;

	case 'h' :
		usage();
		return 0;

	case 'i' :
		stat_interval = atof(optarg) * 1000000000LL;
		break;

	case 'j' :
		if (parse_pattern(optarg))
			fatal("Invalid pattern\n");
		break;

	case 'l' :
		length = atoi(optarg);
		if (length < offsetof(struct message, payload) || length > 65000)
			fatal("Invalid packet length %d\n", length);
		break;

	case 'm':
		mc_group_display = 1;
		break;

	case 'n':
		number_of_multicast_groups = atoi(optarg);
		if (number_of_multicast_groups < 0 || number_of_multicast_groups > 10000)
			fatal("Invalid number of multicast groups\n");
		break;

	case 'o':
		offset_into_pattern = atoi(optarg);
		break;

	case 'p' :
		port = atoi(optarg);
		if (port < 1024 || port > 60000)
			fatal("Invalid port number %d\n", port);
		break;

	case 'r' :
		rate = atoi(optarg);
		if (rate > 1000000 || rate < 1)
			fatal("Invalid rate %d\n", rate);

		wait_time.tv_nsec = 1000000000LL / rate;
		break;

	case 's':
		samples = atoi(optarg);
		if (samples < 0)
			fatal("Invalid number of samples\n");
		break;

	case 't':
		ttl = atoi(optarg);
		break;

	case 'v':
		verbose++;
		break;

	case 'A' :
		aggregate = 1;
		break;

	case 'B' :
		burst = atoi(optarg);
		if (burst < -100000 || burst > 10000 || burst == 0)
			fatal("Invalid burst count\n");
		break;

	case 'R':
		randomize = 1;
		break;
	case 'X':
		terminate_remote = 1;
		break;

	case 'L':
		multicast_loop = 1;
		break;

	case 'S':
		synchronize_systems = atoi(optarg);
		if (synchronize_systems < 1 || synchronize_systems > 100)
			fatal("Invalid number of systems to synchronize\n");
		break;

	default:
		fatal("%s: Invalid option '%c'\n", argv[0], optopt);
	}

	if (argc > 1 + optind) {
		usage();
		return 1;
	}

#ifdef USE_TSC
	ll_init(verbose);
#endif
	find_local_addr();
	reply = calloc(1, length);
	init_message(reply, length);

	if (argc == optind) {
		if (offset_into_pattern + number_of_multicast_groups > max_groups)
			fatal("Pattern does not support that many MC addresses\n");

		start_sender();
		receiver_loop();
	} else {
		printf("Extra parameters\n");
		usage();
		return 1;
	}
	stop_sender();
	return 0;

}

